# -*- coding: utf-8 -*-

import re
import warnings
from copy import deepcopy
from pathlib import Path
import xml.etree.ElementTree as ET
from typing import Any, Optional, Sequence, Dict, List, Union
from numbers import Real

import numpy as np
import pandas as pd
import h5py
import scipy.signal as SS

from ...cfg import DEFAULTS
from ...utils.download import http_get
from ...utils.misc import add_docstring
from ..base import _DataBase, DataBaseInfo


__all__ = [
    "CACHET_CADB",
]


_CACHET_CADB_INFO = DataBaseInfo(  # NOT finished yet
    title="""
    CACHET-CADB: A Contextualized Ambulatory Electrocardiography Arrhythmia Dataset
    """,
    about=r"""
    1. The database has 259 days of contextualized ECG recordings from 24 patients and 1,602 manually annotated 10 s heart-rhythm samples.
    2. The length of the ECG records in the CACHET-CADB varies from 24 h to 3 weeks.
    3. The patient's ambulatory context information (activities, movement acceleration, body position, etc.) is extracted for every 10 s interval cumulatively.
    4. nearly 11% of the ECG data in the database is found to be noisy.
    5. Webpages for downloading the database [1]_ and the short-format database [2]_, see also the GitHub repository [3]_.
    """,
    usage=[
        "ECG arrhythmia detection",
        "Self-Supervised Learning",
    ],
    references=[
        "https://data.dtu.dk/articles/dataset/CACHET-CADB/14547264",
        "https://data.dtu.dk/articles/dataset/CACHET-CADB_Short_Format/14547330",
        "https://github.com/cph-cachet/cachet-ecg-db/",
    ],
    doi=[
        "10.3389/fcvm.2022.893090",
        "10.11583/DTU.14547264",
        "10.11583/DTU.14547330",
    ],
)


@add_docstring(_CACHET_CADB_INFO.format_database_docstring(), mode="prepend")
class CACHET_CADB(_DataBase):
    """
    Parameters
    ----------
    db_dir : str or pathlib.Path, optional
        Storage path of the database.
        If not specified, data will be fetched from Physionet.
    working_dir : str, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Level of logging verbosity.
    kwargs : dict, optional
        Auxilliary key word arguments

    """

    __name__ = "CACHET_CADB"

    def __init__(
        self,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[Union[str, Path]] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            db_name="CACHET-CADB",
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )
        self.data_ext = "ecg.bin"
        self.ann_ext = "annotation.csv"
        self.context_data_ext = {
            "acc": "acc.bin",
            "angularrate": "angularrate.bin",
            "hr_live": "hr_live.bin",
            "hrvrmssd_live": "hrvrmssd_live.bin",
            "marker": "marker.csv",
            "movementacceleration_live": "movementacceleration_live.bin",
            "press": "press.bin",
        }
        self.header_ext = "unisens.xml"
        self.header_keys = [
            "adcResolution",
            "baseline",
            "contentClass",
            "dataType",
            "id",
            "lsbValue",
            "sampleRate",
            "unit",
        ]
        self.context_ann_ext = "context.xlsx"

        self.class_map = {1: "AF", 2: "NSR", 3: "Noise", 4: "Others"}
        self.body_position_map = {
            0: "unknown",
            1: "lying supine",
            2: "lying left",
            3: "lying prone",
            4: "lying right",
            5: "upright",
            6: "sitting/lying",
            7: "standing",
        }
        self.activity_map = {
            0: "unknown",
            1: "lying",
            2: "sitting/standing",
            3: "cycling",
            4: "slope up",
            5: "jogging",
            6: "slope down",
            7: "walking",
            8: "sitting/lying",
            9: "standing",
        }
        self.wear_map = {0: "wear", 1: "non-wear"}

        self.fs = 1024

        self._version = "v1"

        self._subject_pattern = "^P[\\d]{1,2}|PNSR\\-\\d$"
        self._df_metadata = None
        self._metadata_cols = [
            "age",
            "gender",
            "weight",
            "height",
            "comment",
            "duration",
            "measurementId",
            "timestampStart",
            "version",
            "sensorLocation",
            "sensorVersion",
            "sensorType",
            "personId",
        ]
        self._all_records = None
        self._all_subjects = None
        self._subject_records = None
        self._full_data_dir = None
        self._full_ann_dir = None
        self._short_format_file = None
        self.__short_format_data = None
        self.__short_format_ann = None
        self._ls_rec()

    def _ls_rec(self) -> None:
        """Find all records in the database directory
        and store them (path, metadata, etc.) in some private attributes.

        TODO: impelement subsampling for the long format data.
        """
        # short format file
        self._short_format_file = list(self.db_dir.rglob("*.hdf5"))
        if len(self._short_format_file) == 1:
            self._short_format_file = self._short_format_file[0]
            with h5py.File(self._short_format_file, "r") as f:
                self.__short_format_data = f["signal"][:].astype(DEFAULTS.DTYPE.NP)
                self.__short_format_ann = f["labels"][:].astype(int)
        else:
            self._short_format_file = None

        # the whole database
        candidate_folders = [
            f
            for f in list(self.db_dir.rglob("*"))
            if re.match(self._subject_pattern, f.name) is not None
        ]
        self._all_subjects = sorted(set([f.name for f in candidate_folders]))
        if len(self._all_subjects) == 0:
            return  # no record found
        assert set([f.parent.name for f in candidate_folders]) == {
            "annotations",
            "signal",
        }, "invalid folder structure"
        self._full_data_dir = candidate_folders[0].parents[1] / "signal"
        self._full_ann_dir = candidate_folders[0].parents[1] / "annotations"

        self._all_records = []
        for subject in self._all_subjects:
            ann_dir = self._full_ann_dir / subject
            for file_path in ann_dir.rglob(self.ann_ext):
                segment = file_path.parents[0].name
                measurement_id = file_path.parents[1].name
                self._all_records.append(f"{subject}/{measurement_id}/{segment}")
        self._df_records = pd.DataFrame()
        self._df_records["record"] = self._all_records
        self._df_records["subject"] = self._df_records["record"].apply(
            lambda x: x.split("/")[0]
        )
        self._df_records["data_path"] = self._df_records["record"].apply(
            lambda x: self._full_data_dir / f"{x}/{self.data_ext}"
        )
        self._df_records["ann_path"] = self._df_records["record"].apply(
            lambda x: self._full_ann_dir / f"{x}/{self.ann_ext}"
        )
        self._df_records["context_ann_path"] = self._df_records["record"].apply(
            lambda x: self._full_ann_dir / f"{x}/{self.context_ann_ext}"
        )
        for context_name, context_ext in self.context_data_ext.items():
            self._df_records[f"context_{context_name}_path"] = self._df_records[
                "record"
            ].apply(lambda x: self._full_data_dir / f"{x}/{context_ext}")
        self._df_records["header_path"] = self._df_records["record"].apply(
            lambda x: self._full_data_dir / f"{x}/{self.header_ext}"
        )
        self._df_records.set_index("record", inplace=True)
        self._subject_records = {
            subject: sorted(
                self._df_records[self._df_records["subject"] == subject].index
            )
            for subject in self._all_subjects
        }

        # the table of metadata
        self._df_metadata = pd.DataFrame(
            [self.get_record_metadata(rec) for rec in self._all_records]
        )
        self._df_metadata["record"] = self._all_records
        self._df_metadata.set_index("record", inplace=True)
        self._df_metadata = self._df_metadata[self._metadata_cols]

    def get_absolute_path(
        self, rec: Union[str, int], extension: str = "signal-ecg"
    ) -> Path:
        """Get the absolute path of the signal folder of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        extension: str, default "signal-ecg"
            Extension of the file, can be one of
            "header", "annotation", "signal",
            "annotation-context",
            "signal-ecg", "signal-acc", "signal-angularrate",
            "signal-hr_live", "signal-hrvrmssd_live", etc.

        Returns
        -------
        pathlib.Path
            Absolute path of the file.

        """
        if isinstance(rec, int):
            rec = self[rec]
        assert rec in self.all_records, f"invalid record name: `{rec}`"
        if extension == "annotation":
            return self._df_records.loc[rec, "ann_path"]
        elif extension == "annotation-context":
            return self._df_records.loc[rec, "context_ann_path"]
        elif extension == "signal":
            return self._df_records.loc[rec, "data_path"].parent
        elif extension == "header":
            return self._df_records.loc[rec, "header_path"]

        ext1, ext2 = extension.split("-")
        assert ext1 == "signal", f"extension `{extension}` not supported"
        assert ext2 in list(self.context_data_ext) + [
            "ecg"
        ], f"extension `{extension}` not supported"
        if ext2 == "ecg":
            return self._df_records.loc[rec, "data_path"]
        else:
            return self._df_records.loc[rec, f"context_{ext2}_path"]

    def get_subject_id(self, rec: Union[str, int]) -> str:
        """Attach a unique subject ID for the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        sid : str
            Subject ID attached to the record.

        """
        if isinstance(rec, int):
            rec = self[rec]
        assert rec in self.all_records, f"record `{rec}` not found"
        return self._df_records.loc[rec, "subject"]

    def _rdheader(
        self, rec: Union[str, int], key: Optional[str] = None, raw: bool = False
    ) -> Union[str, Dict[str, Any]]:
        """Read header file of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        key : str, optional
            Key of the header to be read, can be one of
            "acc", "angularrate", "hr_live", "hrvrmssd_live",
            "movementacceleration_live", "press", "ecg", "customAttributes".
            If is None, read all keys.
        raw : bool, default False
            If True, return the raw header file,
            otherwise, return a dict.

        Returns
        -------
        header : dict or str
            Header information, or raw header file,
            depending on the value of `raw`.

        """
        if isinstance(rec, int):
            rec = self[rec]
        header_path = self.get_absolute_path(rec, "header")
        if raw:
            return header_path.read_text()
        tree = ET.parse(header_path)
        # return tree
        root = tree.getroot()
        header = {}
        for child in root:
            if "customAttributes" in child.tag:
                header["customAttributes"] = {
                    k: v for k, v in root.attrib.items() if k in self._metadata_cols
                }
                for child2 in child:
                    if child2.attrib["key"] in self._metadata_cols:
                        header["customAttributes"][
                            child2.attrib["key"]
                        ] = child2.attrib["value"]
            elif "signalEntry" in child.tag:
                sig_key = child.attrib["id"].split(".")[0]
                header[sig_key] = deepcopy(child.attrib)
                for child2 in child:
                    if "channel" in child2.tag:
                        if "channel" not in header[sig_key]:
                            header[sig_key]["channel"] = []
                        header[sig_key]["channel"].append(child2.attrib["name"])
        if key is not None:
            assert key in header.keys(), (
                f"key `{key}` not found in header, available keys: `{list(header.keys())}`. "
                f"Perhaps record `{rec}` does not have this type of data?"
            )
            return header[key]
        return header

    def load_data(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        data_format: str = "channel_first",
        units: Union[str, type(None)] = "mV",
        fs: Optional[Real] = None,
    ) -> np.ndarray:
        """Load physical (converted from digital) ECG data,
        or load digital signal directly.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`,
            or "short_format" (-1) to load data from the short format file.
        sampfrom : int, optional
            Start index of the data to be loaded.
        sampto : int, optional
            End index of the data to be loaded.
        data_format : str, default "channel_first"
            Format of the ECG data,
            "channel_last" (alias "lead_last"), or
            "channel_first" (alias "lead_first"), or
            "flat" (alias "plain").
        units : str or None, default "mV"
            Units of the output signal, can also be "μV" (aliases "uV", "muV");
            None for digital data, without digital-to-physical conversion.
        fs : numbers.Real, optional
            Sampling frequency of the output signal.
            If not None, the loaded data will be resampled to this frequency,
            otherwise, the original sampling frequency will be used.

        Returns
        -------
        data : numpy.ndarray
            The loaded ECG data.

        """
        if isinstance(rec, int):
            if rec == -1:
                rec = "short_format"
            else:
                rec = self[rec]
        if rec == "short_format":
            if self._short_format_file is None:
                raise ValueError("Short format file not found")
            return self.__short_format_data
        elif rec not in self.all_records:
            raise ValueError(f"Invalid record name: `{rec}`")
        data_path = self._df_records.loc[rec, "data_path"]
        header = self._rdheader(rec, key="ecg")
        data = np.fromfile(data_path, dtype=header["dataType"])[
            sampfrom or 0 : sampto or None
        ]
        if units is not None:
            # digital to analog conversion using the field `lsbValue` in the header
            data = (data - int(header["baseline"])) * float(header["lsbValue"])
            data = data.astype(DEFAULTS.DTYPE.NP)
            if units.lower() in ["μv", "uv"]:
                data *= 1e3
            elif units.lower() != "mv":
                raise ValueError(f"Invalid `units`: {units}")
        if data_format in ["channel_last", "lead_last"]:
            data = data[:, np.newaxis]
        elif data_format in ["channel_first", "lead_first"]:
            data = data[np.newaxis, :]
        elif data_format not in ["flat", "plain"]:
            raise ValueError(f"Invalid `data_format`: {data_format}")

        if fs is not None and fs != self.fs:
            data = SS.resample_poly(data, fs, self.fs, axis=0).astype(data.dtype)

        return data

    def load_context_data(
        self,
        rec: Union[str, int],
        context_name: str,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        channels: Optional[Union[str, int, List[str], List[int]]] = None,
        units: Optional[str] = None,
        fs: Optional[Real] = None,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Load context data (e.g. accelerometer, heart rate, etc.).

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        context_name : str
            Context name, can be one of
            "acc", "angularrate", "hr_live", "hrvrmssd_live",
            "movementacceleration_live", "press", "marker".
        sampfrom : int, optional
            Start index of the data to be loaded.
        sampto : int, optional
            End index of the data to be loaded.
        channels : str or int or List[str] or List[int], optional
            Channels (names or indices) to be loaded.
            If is None, all channels will be loaded.
        units : str, optional
            Units of the output signal,
            currently can only be "default";
            None for digital data, without digital-to-physical conversion.
        fs : numbers.Real, optional
            Sampling frequency of the output signal.
            If not None, the loaded data will be resampled to this frequency,
            otherwise, the original sampling frequency will be used.

        Returns
        -------
        context_data : numpy.ndarray or pandas.DataFrame
            Context data in the "channel_first" format.

        NOTE
        ----
        If the record does not have the specified context data,
        empty array or DataFrame will be returned.

        """
        if isinstance(rec, int):
            rec = self[rec]
        if rec not in self.all_records:
            raise ValueError(f"Invalid record name: `{rec}`")
        if context_name == "ecg":
            raise ValueError("Call `load_data` to load ECG data")
        assert (
            context_name in self.context_data_ext.keys()
        ), f"Invalid `context_name`: `{context_name}`"
        context_data_path = self._df_records.loc[rec, f"context_{context_name}_path"]

        if context_data_path.suffix == ".csv":
            if not context_data_path.exists():
                warnings.warn(
                    f"record `{rec}` does not have context data `{context_name}`",
                    RuntimeWarning,
                )
                return pd.DataFrame()
            return pd.read_csv(context_data_path, header=None, index_col=None)

        if not context_data_path.exists():
            warnings.warn(
                f"record `{rec}` does not have context data `{context_name}`",
                RuntimeWarning,
            )
            return np.array([])

        header = self._rdheader(rec, key=context_name)
        context_data = np.fromfile(context_data_path, dtype=header["dataType"])[
            sampfrom or 0 : sampto or None
        ]
        if units is not None:
            assert units.lower() in ["default", header["unit"]], (
                f"`units` should be `default` or `{header['unit']}`, but got `{units}`. "
                "Currently, units conversion is not supported."
            )
            # digital to analog conversion using the field `lsbValue` in the header
            context_data = (context_data - int(header.get("baseline", 1))) * float(
                header["lsbValue"]
            )
            context_data = context_data.astype(DEFAULTS.DTYPE.NP)

        # convert to "channel_first" format
        context_data = context_data.reshape(-1, len(header["channel"])).T

        if channels is None:
            return context_data
        _input_channels = channels
        if isinstance(channels, str):
            assert (
                channels in header["channel"]
            ), f"`channels` should be a subset of `{header['channel']}`, but got `{channels}`"
            channels = [header["channel"].index(channels)]
        elif isinstance(channels, int):
            assert channels < len(
                header["channel"]
            ), f"`channels` should be less than `{len(header['channel'])}`, but got `{channels}`"
            channels = [channels]
        else:
            assert all(
                [
                    ch in header["channel"]
                    if isinstance(ch, str)
                    else ch in range(len(header["channel"]))
                    for ch in channels
                ]
            ), f"`channels` should be a subset of `{header['channel']}`, but got `{_input_channels}`"
            _channels = [
                header["channel"].index(ch) if isinstance(ch, str) else ch
                for ch in channels
            ]
            channels = list(dict.fromkeys(_channels))
            if len(channels) != len(_channels):
                warnings.warn(
                    f"duplicate `channels` are removed, {_input_channels} -> {channels}",
                    RuntimeWarning,
                )
        context_data = context_data[channels]

        if fs is not None and fs != header["sampleRate"]:
            context_data = SS.resample_poly(
                context_data, fs, header["sampleRate"], axis=1
            ).astype(context_data.dtype)

        return context_data

    def load_ann(
        self, rec: Union[str, int], ann_format: str = "pd"
    ) -> Union[pd.DataFrame, np.ndarray, Dict[Union[int, str], np.ndarray]]:
        """Load annotation from the metadata file.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        ann_format : str, default "pd"
            Format of the annotation,
            currently only "pd" is supported.

        Returns
        -------
        ann : pandas.DataFrame or numpy.ndarray or dict
            The annotation of the record.

        """
        if isinstance(rec, int):
            if rec == -1:
                rec = "short_format"
            else:
                rec = self[rec]
        if rec == "short_format":
            if self._short_format_file is None:
                raise ValueError("Short format file not found")
            return self.__short_format_ann
        elif rec not in self.all_records:
            raise ValueError(f"Invalid record name: `{rec}`")
        try:
            ann = pd.read_csv(self._df_records.loc[rec, "ann_path"])
        except pd.errors.EmptyDataError:
            ann = pd.DataFrame(columns=["Start", "End", "Class"])
        # TODO: adjust return format according to `ann_format`
        if ann_format == "pd":
            return ann
        else:
            raise ValueError(f"`ann_format`: `{ann_format}` not supported")

    def load_context_ann(
        self, rec: Union[str, int], sheet_name: Optional[str] = None
    ) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
        """Load context annotation.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sheet_name : str, optional
            Sheet name of the context annotation file, can be one of
            "movisens DataAnalyzer Parameter", "movisens DataAnalyzer Results".
            If is None, all sheets will be loaded.

        Returns
        -------
        context_ann : pandas.DataFrame or dict
            Context annotations of the record.

        """
        if isinstance(rec, int):
            rec = self[rec]
        if rec not in self.all_records:
            raise ValueError(f"Invalid record name: `{rec}`")
        context_ann_path = self._df_records.loc[rec, "context_ann_path"]
        context_ann = pd.read_excel(
            context_ann_path, engine="openpyxl", sheet_name=sheet_name
        )
        return context_ann

    def get_record_metadata(self, rec: Union[str, int]) -> Dict[str, str]:
        """Get metadata of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`,
            or "short_format" (-1) to load data from the short format file.

        Returns
        -------
        metadata : dict
            Metadata of the record

        """
        metadata = self._rdheader(rec, key="customAttributes")
        return metadata

    def get_subject_info(
        self, rec_or_sid: Union[str, int], items: Optional[List[str]] = None
    ) -> Dict[str, str]:
        """Read auxiliary information of a subject (a record)
        stored in the header files.

        Parameters
        ----------
        rec : str or int
            Record name, or index of the record in :attr:`all_records`,
            or the subject ID.
        items : List[str], optional
            Items of the subject"s information (e.g. sex, age, etc.).

        Returns
        -------
        subject_info : dict
            Information about the subject, including
            "age", "gender", "height", "weight".

        """
        subject_info_items = ["age", "gender", "height", "weight"]
        if rec_or_sid in self.all_subjects:
            rec = self.subject_records[rec_or_sid][0]
        elif isinstance(rec_or_sid, int):
            rec = self[rec_or_sid]
        else:
            rec = rec_or_sid
        assert rec in self.all_records, f"invalid record name: `{rec}`"
        subject_info = self._rdheader(rec, key="customAttributes")
        if items is not None:
            assert set(items).issubset(list(subject_info)), f"invalid items: `{items}`"
            subject_info = {k: subject_info[k] for k in items}
        else:
            subject_info = {k: subject_info[k] for k in subject_info_items}
        return subject_info

    @property
    def all_subjects(self) -> List[str]:
        """List of all subject IDs."""
        return self._all_subjects

    @property
    def subject_records(self) -> Dict[str, List[str]]:
        """Dict of subject IDs and their corresponding records."""
        return self._subject_records

    @property
    def df_metadata(self) -> pd.DataFrame:
        """The table of metadata of the records."""
        return self._df_metadata

    @property
    def url(self) -> Dict[str, str]:
        return {
            "CACHET-CADB.zip": "https://data.dtu.dk/ndownloader/files/27928830",
            "cachet-cadb_short_format_without_context.hdf5.zip": "https://data.dtu.dk/ndownloader/files/27917358",
        }

    def download(self, files: Optional[Union[str, Sequence[str]]]) -> None:
        """Download the database from the DTU website.

        Parameters
        ----------
        files : str or Sequence[str], optional
            Files to download, can be subset of
            "CACHET-CADB.zip", "cachet-cadb_short_format_without_context.hdf5.zip".
            If is None, all files will be downloaded.

        """
        warnings.warn(
            "The files are large, and the connections are unstable. "
            "One might need some downloading tools (e.g. Xunlei) to download the files.",
            RuntimeWarning,
        )
        if files is None:
            files = self.url.keys()
        if isinstance(files, str):
            files = [files]
        assert set(files).issubset(
            self.url
        ), f"`files` should be a subset of `{list(self.url)}`"
        for filename in files:
            url = self.url[filename]
            if not (self.db_dir / filename).is_file():
                http_get(url, self.db_dir, filename=filename)
        self._ls_rec()

    def plot(
        self,
        rec: Union[str, int],
        **kwargs: Any,
    ) -> None:
        """Not implemented."""
        raise NotImplementedError

    @property
    def database_info(self) -> DataBaseInfo:
        return _CACHET_CADB_INFO



import json
import textwrap
import time
import warnings
from copy import deepcopy
from random import sample, shuffle
from typing import List, Optional, Sequence, Set, Tuple, Any

import numpy as np
import torch
from torch.utils.data.dataset import Dataset
from tqdm.auto import tqdm

from ...._preprocessors import PreprocManager
from ....cfg import CFG
from ....databases import CINC2021 as CR
from ....utils.ecg_arrhythmia_knowledge import Standard12Leads
from ....utils.misc import ReprMixin, list_sum
from ....utils.utils_data import ensure_siglen
from ....utils.utils_signal import remove_spikes_naive


__all__ = [
    "CINC2021Dataset",
]


class CINC2021Dataset(ReprMixin, Dataset):
    """Data generator for feeding data into pytorch models
    using the :class:`~torch_ecg.databases.CINC2021` database.

    Parameters
    ----------
    config : dict
        configurations for the :class:`Dataset`,
        ref. `CINC2021TrainCfg`.
        A simple example is as follows:

        .. code-block:: python

            >>> config = deepcopy(CINC2021TrainCfg)
            >>> config.db_dir = "some/path/to/db"
            >>> dataset = CINC2021Dataset(config, training=True, lazy=False)

    training : bool, default True
        If True, the training set will be loaded,
        otherwise the test (val) set will be loaded.
    lazy : bool, default True
        If True, the data will not be loaded immediately,
        instead, it will be loaded on demand.
    **reader_kwargs : dict, optional
        Keyword arguments for the database reader class.

    """

    __DEBUG__ = False
    __name__ = "CINC2021Dataset"

    def __init__(
        self,
        config: Optional[CFG] = None,
        training: bool = True,
        lazy: bool = True,
        **reader_kwargs: Any,
    ) -> None:
        super().__init__()
        self.config = deepcopy(config)
        if reader_kwargs.pop("db_dir", None) is not None:
            warnings.warn(
                "`db_dir` is specified in both config and reader_kwargs", RuntimeWarning
            )
        self.reader = CR(db_dir=self.config.db_dir, **reader_kwargs)
        # assert self.config.db_dir is not None, "db_dir must be specified"
        self.config.db_dir = self.reader.db_dir
        self._TRANCHES = (
            self.config.tranche_classes.keys()
        )  # ["A", "B", "AB", "E", "F", "G",]
        self.tranches = self.config.tranches_for_training
        self.training = training
        self.dtype = self.config.np_dtype
        assert not self.tranches or self.tranches in self._TRANCHES
        if self.tranches:
            self.all_classes = self.config.tranche_classes[self.tranches]
            self.class_weights = self.config.tranche_class_weights[self.tranches]
        else:
            self.all_classes = self.config.classes
            self.class_weights = self.config.class_weights
        self.config.all_classes = deepcopy(self.all_classes)
        self.n_classes = len(self.all_classes)
        # print(f"tranches = {self.tranches}, all_classes = {self.all_classes}")
        # print(f"class_weights = {dict_to_str(self.class_weights)}")
        cw = np.zeros((len(self.class_weights),), dtype=self.dtype)
        for idx, c in enumerate(self.all_classes):
            cw[idx] = self.class_weights[c]
        self.class_weights = torch.from_numpy(cw.astype(self.dtype)).view(
            1, self.n_classes
        )
        # validation also goes in batches, hence length has to be fixed
        self.siglen = self.config.input_len
        self.lazy = lazy

        self._indices = [Standard12Leads.index(ld) for ld in self.config.leads]

        self.records = self._train_test_split(
            self.config.train_ratio, force_recompute=False
        )
        # TODO: consider using `remove_spikes_naive` to treat these exceptional records
        self.records = [
            r
            for r in self.records
            if r not in self.reader.exceptional_records
            and (self.reader.get_data_filepath(r)).is_file()
        ]
        if self.__DEBUG__:
            self.records = sample(self.records, int(len(self.records) * 0.01))

        ppm_config = CFG(random=False)
        ppm_config.update(self.config)
        self.ppm = PreprocManager.from_config(ppm_config)
        # self.ppm.rearrange(["bandpass", "normalize"])

        self._signals = np.array([], dtype=self.dtype).reshape(
            0, len(self.config.leads), self.siglen
        )
        self._labels = np.array([], dtype=self.dtype).reshape(0, self.n_classes)
        if not self.lazy:
            self._load_all_data()

    def _load_all_data(self) -> None:
        """Load all data into memory."""
        fdr = _FastDataReader(self.reader, self.records, self.config, self.ppm)

        # with tqdm(self.records, desc="Loading data", unit="record") as pbar:
        #     for rec in pbar:
        # s, l = self._load_one_record(rec)  # self._load_one_record is much slower than _FastDataReader
        self._signals, self._labels = [], []
        with tqdm(
            desc="Loading data",
            total=len(fdr),
            unit="record",
            dynamic_ncols=True,
            mininterval=1.0,
        ) as pbar:
            for idx in range(len(fdr)):
                sig, lb = fdr[idx]
                # np.concatenate slows down the process severely
                # self._signals = np.concatenate((self._signals, s), axis=0)
                # self._labels = np.concatenate((self._labels, l), axis=0)
                self._signals.append(sig)
                self._labels.append(lb)
                pbar.update(1)
        self._signals = np.concatenate(self._signals, axis=0).astype(self.dtype)
        self._labels = np.concatenate(self._labels, axis=0)

    def _load_one_record(self, rec: str) -> Tuple[np.ndarray, np.ndarray]:
        """Load one record from the database using data reader.

        NOTE
        ----
        DO NOT USE THIS FUNCTION DIRECTLY for preloading data,
        use `_FastDataReader` instead.

        Parameters
        ----------
        rec : str
            Name of the record.

        Returns
        -------
        values : numpy.ndarray
            The signal values of the record.
        labels : numpy.ndarray
            The labels of the record.

        """
        values = self.reader.load_resampled_data(
            rec,
            leads=self.config.leads,
            # leads=Standard12Leads,
            data_format=self.config.data_format,
            siglen=None,
        )
        for idx in range(values.shape[0]):
            values[idx] = remove_spikes_naive(values[idx])
        values, _ = self.ppm(values, self.config.fs)
        values = ensure_siglen(
            values,
            siglen=self.siglen,
            fmt=self.config.data_format,
            tolerance=self.config.sig_slice_tol,
        ).astype(self.dtype)
        if values.ndim == 2:
            values = values[np.newaxis, ...]

        labels = self.reader.get_labels(rec, scored_only=True, fmt="a", normalize=True)
        labels = (
            np.isin(self.all_classes, labels)
            .astype(self.dtype)[np.newaxis, ...]
            .repeat(values.shape[0], axis=0)
        )

        return values, labels

    def to(self, leads: Sequence[str]) -> None:
        """Reduce the leads to a subset of the current leads.

        Parameters
        ----------
        leads : Sequence[str]
            The subset of leads to be kept.

        Returns
        -------
        None

        """
        assert set(leads) <= set(
            self.config.leads
        ), "One is not able to change to a set of leads which is not a subset of the current leads"
        prev_leads = self.config.leads
        self.config.leads = leads
        self._indices = [prev_leads.index(ld) for ld in leads]
        self._signals = self._signals[:, self._indices, :]

    def empty(self, leads: Optional[Sequence[str]] = None) -> None:
        """Empty the loaded data, and keep the leads.

        Parameters
        ----------
        leads : Sequence[str], optional
            The leads to be kept, by default the current leads.

        Returns
        -------
        None

        """
        if leads is None:
            leads = self.config.leads
        else:
            self.config.leads = leads
        self._signals = np.array([], dtype=self.dtype).reshape(
            0, len(leads), self.siglen
        )
        self._labels = np.array([], dtype=self.dtype).reshape(0, self.n_classes)

    @classmethod
    def from_extern(cls, ext_ds: "CINC2021Dataset", config: CFG) -> "CINC2021Dataset":
        """Create a new dataset from an existing one.

        The existing dataset must have all data loaded into memory,
        and its leads must be a superset of the new dataset.

        Parameters
        ----------
        ext_ds : CINC2021Dataset
            The existing dataset.
        config : CFG
            The configuration of the new dataset.

        Returns
        -------
        CINC2021Dataset
            A new instance of the dataset.

        """
        new_ds = cls(config, ext_ds.training, lazy=True)
        indices = [ext_ds.config.leads.index(ld) for ld in new_ds.config.leads]
        new_ds._signals = ext_ds._signals[:, indices, :]
        new_ds._labels = ext_ds._labels.copy()
        return new_ds

    def reload_from_extern(self, ext_ds: "CINC2021Dataset") -> None:
        """Reload the data from an existing dataset.

        The existing dataset must have all data loaded into memory,
        and its leads must be a superset of the current dataset.

        Parameters
        ----------
        ext_ds : CINC2021Dataset
            The existing dataset.

        Returns
        -------
        None

        """
        assert set(self.config.leads) <= set(
            ext_ds.config.leads
        ), "One is not able to reload from a dataset whose `leads` is not a superset of the current one"
        indices = [ext_ds.config.leads.index(ld) for ld in self.config.leads]
        self._signals = ext_ds._signals[:, indices, :]
        self._labels = ext_ds._labels.copy()

    @property
    def signals(self) -> np.ndarray:
        """Cached signals, only available when `lazy=False`
        or preloading is performed manually.
        """
        return self._signals

    @property
    def labels(self) -> np.ndarray:
        """Cached labels, only available when `lazy=False`
        or preloading is performed manually.
        """
        return self._labels

    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        return self.signals[index], self.labels[index]

    def __len__(self) -> int:
        return len(self._signals)

    def _train_test_split(
        self,
        train_ratio: float = 0.8,
        force_recompute: bool = False,
        force_valid: bool = False,
    ) -> List[str]:
        """Perform train-test split.

        It is ensured that both the train and the test set
        contain all classes if `force_valid` is True.

        Parameters
        ----------
        train_ratio : float, default 0.8
            Ratio of the train set in the whole dataset
            (or the whole tranche(s)).
        force_recompute : bool, default False
            If True, the train-test split will be recomputed,
            regardless of the existing ones stored in json files.
        force_valid : bool, default False
            If True, the train-test split will be recomputed
            if validity check fails.

        Returns
        -------
        records : List[str]
            List of the records split for training
            or for testing (validation).

        """
        time.sleep(1)
        start = time.time()
        print("\nstart performing train test split...\n")
        time.sleep(1)
        _TRANCHES = list("ABEFG")
        _train_ratio = int(train_ratio * 100)
        _test_ratio = 100 - _train_ratio
        assert _train_ratio * _test_ratio > 0

        ns = "_ns" if len(self.config.special_classes) == 0 else ""
        file_suffix = f"_siglen_{self.siglen}{ns}.json"
        train_file = (
            self.reader.db_dir_base
            / f"{self.reader.db_name}_train_ratio_{_train_ratio}{file_suffix}"
        )
        test_file = (
            self.reader.db_dir_base
            / f"{self.reader.db_name}_test_ratio_{_test_ratio}{file_suffix}"
        )

        # TODO: use self.reader.df_stats (precomputed and stored in utils/stats.csv)
        # to accelerate the validity examinations
        if force_recompute or not all([train_file.is_file(), test_file.is_file()]):
            tranche_records = {t: [] for t in _TRANCHES}
            train_set = {t: [] for t in _TRANCHES}
            test_set = {t: [] for t in _TRANCHES}
            for t in _TRANCHES:
                with tqdm(
                    self.reader.all_records[t],
                    total=len(self.reader.all_records[t]),
                    dynamic_ncols=True,
                    mininterval=1.0,
                ) as bar:
                    for rec in bar:
                        if rec in self.reader.exceptional_records:
                            # skip exceptional records
                            continue
                        rec_labels = self.reader.get_labels(
                            rec, scored_only=True, fmt="a", normalize=True
                        )
                        rec_labels = [
                            c for c in rec_labels if c in self.config.tranche_classes[t]
                        ]
                        if len(rec_labels) == 0:
                            # skip records with no scored class
                            continue
                        rec_samples = self.reader.load_resampled_data(rec).shape[1]
                        # NEW in CinC2021 compared to CinC2020
                        # training input siglen raised from 4000 to 5000,
                        # hence allow tolerance in siglen now
                        if rec_samples < self.siglen - self.config.input_len_tol:
                            continue
                        tranche_records[t].append(rec)
                time.sleep(1)
                print(
                    f"tranche {t} has {len(tranche_records[t])} valid records for training"
                )
            for t in _TRANCHES:
                is_valid = False
                while not is_valid:
                    shuffle(tranche_records[t])
                    split_idx = int(len(tranche_records[t]) * train_ratio)
                    train_set[t] = tranche_records[t][:split_idx]
                    test_set[t] = tranche_records[t][split_idx:]
                    if force_valid:
                        is_valid = self._check_train_test_split_validity(
                            train_set[t],
                            test_set[t],
                            set(self.config.tranche_classes[t]),
                        )
                    else:
                        is_valid = True
            train_file.write_text(json.dumps(train_set, ensure_ascii=False))
            test_file.write_text(json.dumps(test_set, ensure_ascii=False))
            print(
                textwrap.dedent(
                    f"""
                train set saved to \n\042{str(train_file)}\042
                test set saved to \n\042{str(test_file)}\042
                """
                )
            )
        else:
            train_set = json.loads(train_file.read_text())
            test_set = json.loads(test_file.read_text())

        print(f"train test split finished in {(time.time()-start)/60:.2f} minutes")

        _tranches = list(self.tranches or "ABEFG")
        if self.training == "all":
            records = list_sum([train_set[k] for k in _tranches]) + list_sum(
                [test_set[k] for k in _tranches]
            )
        elif self.training is True:
            records = list_sum([train_set[k] for k in _tranches])
        else:
            records = list_sum([test_set[k] for k in _tranches])
        return records

    def _check_train_test_split_validity(
        self, train_set: List[str], test_set: List[str], all_classes: Set[str]
    ) -> bool:
        """Check if the train-test split is valid.

        The train-test split is valid iff
        records in both `train_set` and `test` contain all classes in `all_classes`

        Parameters
        ----------
        train_set : List[str]
            List of the records in the train set.
        test_set : List[str]
            List of the records in the test set.
        all_classes : Set[str]
            The set of all classes for training.

        Returns
        -------
        is_valid : bool
            Whether the train-test split is valid or not.

        """
        train_classes = set(
            list_sum([self.reader.get_labels(rec, fmt="a") for rec in train_set])
        )
        train_classes.intersection_update(all_classes)
        test_classes = set(
            list_sum([self.reader.get_labels(rec, fmt="a") for rec in test_set])
        )
        test_classes.intersection_update(all_classes)
        is_valid = len(all_classes) == len(train_classes) == len(test_classes)
        print(
            textwrap.dedent(
                f"""
            all_classes:     {all_classes}
            train_classes:   {train_classes}
            test_classes:    {test_classes}
            is_valid:        {is_valid}
            """
            )
        )
        return is_valid

    def persistence(self) -> None:
        """Save the processed dataset to disk."""
        _TRANCHES = "ABEFG"
        if self.training:
            ratio = int(self.config.train_ratio * 100)
        else:
            ratio = 100 - int(self.config.train_ratio * 100)
        fn_suffix = f"tranches_{self.tranches or _TRANCHES}_ratio_{ratio}"
        if self.config.bandpass is not None:
            bp_low = max(0, self.config.bandpass[0])
            bp_high = min(self.config.bandpass[1], self.config.fs // 2)
            fn_suffix = fn_suffix + f"_bp_{bp_low:.1f}_{bp_high:.1f}"
        fn_suffix = fn_suffix + f"_siglen_{self.siglen}"

        X, y = [], []
        with tqdm(
            range(self.__len__()),
            total=self.__len__(),
            dynamic_ncols=True,
            mininterval=1.0,
        ) as bar:
            for idx in bar:
                values, labels = self.__getitem__(idx)
                X.append(values)
                y.append(labels)
        X, y = np.array(X), np.array(y)
        print(f"X.shape = {X.shape}, y.shape = {y.shape}")
        filename = f"{'train' if self.training else 'test'}_X_{fn_suffix}.npy"
        np.save(self.reader.db_dir_base / filename, X)
        print(f"X saved to {filename}")
        filename = f"{'train' if self.training else 'test'}_y_{fn_suffix}.npy"
        np.save(self.reader.db_dir_base / filename, y)
        print(f"y saved to {filename}")

    def _check_nan(self) -> None:
        """Check if there are nan values in the dataset."""
        for idx, (values, labels) in enumerate(self):
            if np.isnan(values).any():
                print(f"values of {self.records[idx]} have nan values")
            if np.isnan(labels).any():
                print(f"labels of {self.records[idx]} have nan values")

    def extra_repr_keys(self) -> List[str]:
        return [
            "training",
            "tranches",
            "reader",
        ]


class _FastDataReader(ReprMixin, Dataset):
    """Fast data reader.

    Parameters
    ----------
    reader : CR
        The reader to read the data.
    records : Sequence[str]
        The list of records to read.
    config : CFG
        The configuration.
    ppm : PreprocManager, optional
        The preprocessor manager.

    """

    def __init__(
        self,
        reader: CR,
        records: Sequence[str],
        config: CFG,
        ppm: Optional[PreprocManager] = None,
    ) -> None:
        self.reader = reader
        self.records = records
        self.config = config
        self.ppm = ppm
        self.dtype = self.config.np_dtype

    def __len__(self) -> int:
        return len(self.records)

    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        rec = self.records[index]
        values = self.reader.load_resampled_data(
            rec,
            leads=self.config.leads,
            # leads=Standard12Leads,
            data_format=self.config.data_format,
            siglen=None,
        )
        for idx in range(values.shape[0]):
            values[idx] = remove_spikes_naive(values[idx])
        if self.ppm:
            values, _ = self.ppm(values, self.config.fs)
        values = ensure_siglen(
            values,
            siglen=self.config.input_len,
            fmt=self.config.data_format,
            tolerance=self.config.sig_slice_tol,
        ).astype(self.dtype)
        if values.ndim == 2:
            values = values[np.newaxis, ...]

        labels = self.reader.get_labels(rec, scored_only=True, fmt="a", normalize=True)
        labels = (
            np.isin(self.config.all_classes, labels)
            .astype(self.dtype)[np.newaxis, ...]
            .repeat(values.shape[0], axis=0)
        )

        return values, labels

    def extra_repr_keys(self) -> List[str]:
        return [
            "reader",
            "ppm",
        ]



"""
"""

from numbers import Real
from random import choice, randint
from typing import Any, List, Optional, Sequence, Tuple, Union

import numpy as np
import scipy.signal as SS
import torch
import torch.nn.functional as F
from torch import Tensor

from ..cfg import DEFAULTS
from ..utils.misc import ReprMixin, add_docstring
from .base import Augmenter


__all__ = [
    "StretchCompress",
    "StretchCompressOffline",
]


class StretchCompress(Augmenter):
    """Stretch-or-compress augmenter on ECG tensors.

    Rescaling the ECGs by a factor sampled from a normal distribution
    along the time axis.

    Parameters
    ----------
    ratio : numbers.Real, default 6
        Mean ratio of the stretch or compress.
        If it is in the interval[1, 100],
        then it will be transformed to [0, 1].
        The ratio of one batch element is sampled from a normal distribution.
    prob : float, default 0.5
        Probability of the augmenter to be applied.
    inplace : bool, default True
        If True, the input ECGs will be modified inplace.
    kwargs : dict, optional
        Additional keyword arguments.

    Example
    -------
    .. code-block:: python

        sc = StretchCompress()
        sig = torch.randn((32, 12, 5000))
        labels = torch.randint(0, 2, (32, 5000, 26))
        label = torch.randint(0, 2, (32, 26), dtype=torch.float32)
        mask = torch.randint(0, 2, (32, 5000, 3), dtype=torch.float32)
        sig, label, mask = sc(sig, label, mask)

    """

    __name__ = "StretchCompress"

    def __init__(
        self, ratio: Real = 6, prob: float = 0.5, inplace: bool = True, **kwargs: Any
    ) -> None:
        super().__init__()
        self.prob = prob
        assert 0 <= self.prob <= 1, "Probability must be between 0 and 1"
        self.inplace = inplace
        self.ratio = ratio
        if self.ratio > 1:
            self.ratio = self.ratio / 100
        assert (
            0 <= self.ratio <= 1
        ), "Ratio must be between 0 and 1, or between 0 and 100"

    def forward(
        self, sig: Tensor, *labels: Optional[Sequence[Tensor]], **kwargs: Any
    ) -> Tuple[Tensor, ...]:
        """Forward method of the augmenter.

        Parameters
        ----------
        sig : torch.Tensor
            Batched ECGs to be stretched or compressed,
            of shape ``(batch, lead, siglen)``.
        labels : Sequence[torch.Tensor], optional
            Label tensors of the ECGs,
            If set, labels of ``ndim = 3``, of shape ``(batch, label_len, channels)``
            will be stretched or compressed.
            `siglen` should be divisible by `label_len`.
        kwargs : dict, optional
            Not used, but kept for consistency with other augmenters.

        Returns
        -------
        sig : torch.Tensor
            The stretched or compressed ECG tensors.
        labels : Sequence[torch.Tensor], optional
            The stretched or compressed label tensors.

        """
        batch, lead, siglen = sig.shape
        if not self.inplace:
            sig = sig.clone()
        labels = [label.clone() for label in labels]
        if self.prob == 0:
            return (sig, *labels)
        label_len = []
        n_labels = len(labels)
        for idx in range(n_labels):
            if labels[idx].ndim < 3:
                label_len.append(None)
                continue
            labels[idx] = labels[idx].permute(
                0, 2, 1
            )  # (batch, label_len, n_classes) -> (batch, n_classes, label_len)
            ll = labels[idx].shape[-1]
            if ll != siglen:
                labels[idx] = F.interpolate(
                    labels[idx], size=(siglen,), mode="linear", align_corners=True
                )
            label_len.append(ll)
        for batch_idx in self.get_indices(prob=self.prob, pop_size=batch):
            sign = choice([-1, 1])
            ratio = self._sample_ratio()
            # print(f"batch_idx = {batch_idx}, sign = {sign}, ratio = {ratio}")
            new_len = int(round((1 + sign * ratio) * siglen))
            diff_len = abs(new_len - siglen)
            half_diff_len = diff_len // 2
            if sign > 0:  # stretch and cut
                sig[batch_idx, ...] = F.interpolate(
                    sig[batch_idx, ...].unsqueeze(0),
                    size=new_len,
                    mode="linear",
                    align_corners=True,
                )[..., half_diff_len : siglen + half_diff_len].squeeze(0)
                for idx in range(n_labels):
                    if labels[idx].ndim < 3:
                        continue
                    labels[idx][batch_idx, ...] = F.interpolate(
                        labels[idx][batch_idx, ...].unsqueeze(0),
                        size=new_len,
                        mode="linear",
                        align_corners=True,
                    )[..., half_diff_len : siglen + half_diff_len].squeeze(0)
            else:  # compress and pad
                sig[batch_idx, ...] = F.pad(
                    F.interpolate(
                        sig[batch_idx, ...].unsqueeze(0),
                        size=new_len,
                        mode="linear",
                        align_corners=True,
                    ),
                    pad=(half_diff_len, diff_len - half_diff_len),
                    mode="constant",
                    value=0.0,
                ).squeeze(0)
                for idx in range(n_labels):
                    if labels[idx].ndim < 3:
                        continue
                    labels[idx][batch_idx, ...] = F.pad(
                        F.interpolate(
                            labels[idx][batch_idx, ...].unsqueeze(0),
                            size=new_len,
                            mode="linear",
                            align_corners=True,
                        ),
                        pad=(half_diff_len, diff_len - half_diff_len),
                        mode="constant",
                        value=0.0,
                    ).squeeze(0)
        for idx, (label, ll) in enumerate(zip(labels, label_len)):
            if labels[idx].ndim < 3:
                continue
            if ll != siglen:
                labels[idx] = F.interpolate(
                    label, size=(ll,), mode="linear", align_corners=True
                )
            labels[idx] = labels[idx].permute(
                0, 2, 1
            )  # (batch, n_classes, label_len) -> (batch, label_len, n_classes)
        return (sig, *labels)

    def _sample_ratio(self) -> float:
        """Sample the ratio of stretching or compressing."""
        return np.clip(
            DEFAULTS.RNG.normal(self.ratio, 0.382 * self.ratio), 0, 2 * self.ratio
        )

    def _generate(
        self, sig: Tensor, *labels: Optional[Sequence[Tensor]]
    ) -> Union[Tuple[Tensor, ...], Tensor]:
        """NOT finished, NOT checked,

        parallel version of `self.generate`, NOT tested yet!

        Parameters
        ----------
        sig : torch.Tensor
            Batched ECGs to be stretched or compressed,
            of shape ``(batch, lead, siglen)``.
        labels : Sequence[torch.Tensor], optional
            Label tensors of the ECGs.
            If set, should be of ``ndim = 3``,
            and of shapes ``(batch, label_len, n_classes)``.
            `siglen` should be divisible by `label_len`.

        Returns
        -------
        sig : torch.Tensor
            The stretched or compressed ECG tensors.
        labels : Sequence[torch.Tensor], optional
            The stretched or compressed label tensors.

        """
        batch, lead, siglen = sig.shape
        if not self.inplace:
            sig = sig.clone()
        if self.prob == 0:
            if len(labels) == 0:
                return sig
            return (sig, *labels)
        indices = self.get_indices(prob=self.prob, pop_size=batch)
        # with tmp.Pool(processes=4) as pool:
        #     data = pool.starmap(
        #         func=_stretch_compress_one_batch_element,
        #         iterable=[
        #             (
        #                 self.ratio,
        #                 sig[batch_idx, ...].unsqueeze(0),
        #                 *(label[batch_idx, ...].unsqueeze(0) for label in labels)
        #             )
        #             for batch_idx in indices
        #         ],
        #     )
        # for batch_idx, (sig_, *labels_) in zip(indices, data):
        #     sig[batch_idx, ...] = sig_
        #     for idx, label in enumerate(labels_):
        #         labels[idx][batch_idx, ...] = label
        for batch_idx in indices:
            data = _stretch_compress_one_batch_element(
                self.ratio,
                sig[batch_idx, ...].unsqueeze(0),
                *(label[batch_idx, ...].unsqueeze(0) for label in labels),
            )
            sig[batch_idx, ...] = data[0]
            for idx, label in enumerate(data[1:]):
                labels[idx][batch_idx, ...] = label
        if len(labels) == 0:
            return sig
        return (sig, *labels)

    def extra_repr_keys(self) -> List[str]:
        return [
            "ratio",
            "prob",
            "inplace",
        ] + super().extra_repr_keys()


def _stretch_compress_one_batch_element(
    ratio: Real, sig: Tensor, *labels: Sequence[Tensor]
) -> Tensor:
    """Stretch or compress one batch element of the ECGs.

    Parameters
    ----------
    ratio : numbers.Real
        Ratio of the stretch/compress.
    sig : torch.Tensor
        The ECGs to be stretched or compressed,
        of shape ``(1, lead, siglen)``.
    labels : Sequence[torch.Tensor], optional
        Label tensors of the ECGs.
        If set, each should be of ``ndim = 3``,
        and of shape ``(1, label_len, channels)``.
        ``siglen`` should be divisible by ``label_len``.

    Returns
    -------
    sig : torch.Tensor
        The stretched or compressed ECG tensor,
        of shape ``(lead, siglen)``.
    labels : Sequence[torch.Tensor], optional
        The stretched or compressed label tensors,
        of shapes ``(label_len, channels)``.

    """
    labels = list(labels)
    label_len = []
    n_labels = len(labels)
    siglen = sig.shape[-1]
    for idx in range(n_labels):
        if labels[idx].ndim < 3:
            label_len.append(0)
            continue
        labels[idx] = labels[idx].permute(
            0, 2, 1
        )  # (1, label_len, n_classes) -> (1, n_classes, label_len)
        ll = labels[idx].shape[-1]
        if ll != siglen:
            labels[idx] = F.interpolate(
                labels[idx], size=(siglen,), mode="linear", align_corners=True
            )
        label_len.append(ll)
    sign = choice([-1, 1])
    ratio = np.clip(DEFAULTS.RNG.normal(ratio, 0.382 * ratio), 0, 2 * ratio)
    # print(f"batch_idx = {batch_idx}, sign = {sign}, ratio = {ratio}")
    new_len = int(round((1 + sign * ratio) * siglen))
    diff_len = abs(new_len - siglen)
    half_diff_len = diff_len // 2
    if sign > 0:  # stretch and cut
        sig = F.interpolate(sig, size=new_len, mode="linear", align_corners=True,)[
            ..., half_diff_len : siglen + half_diff_len
        ].squeeze(0)
        for idx in range(n_labels):
            if label_len[idx] == 0:
                continue
            labels[idx] = F.interpolate(
                labels[idx],
                size=new_len,
                mode="linear",
                align_corners=True,
            )[..., half_diff_len : siglen + half_diff_len]
    else:  # compress and pad
        sig = F.pad(
            F.interpolate(
                sig,
                size=new_len,
                mode="linear",
                align_corners=True,
            ),
            pad=(half_diff_len, diff_len - half_diff_len),
            mode="constant",
            value=0.0,
        ).squeeze(0)
        for idx in range(n_labels):
            if label_len[idx] == 0:
                continue
            labels[idx] = F.pad(
                F.interpolate(
                    labels[idx],
                    size=new_len,
                    mode="linear",
                    align_corners=True,
                ),
                pad=(half_diff_len, diff_len - half_diff_len),
                mode="constant",
                value=0.0,
            )
    for idx, (label, ll) in enumerate(zip(labels, label_len)):
        if ll == 0:
            labels[idx] = labels[idx].squeeze(0)
            continue
        if ll != siglen:
            labels[idx] = F.interpolate(
                label, size=(ll,), mode="linear", align_corners=True
            )
        labels[idx] = (
            labels[idx].squeeze(0).permute(1, 0)
        )  # (n_classes, label_len) -> (label_len, n_classes)
    if len(labels) > 0:
        return (sig, *labels)
    return sig


class StretchCompressOffline(ReprMixin):
    """Offline stretch-or-compress augmenter.

    Stretch-or-compress augmenter on
    orginal length-varying ECG signals (in the form of numpy arrays),
    for the purpose of offline data generation.

    Parameters
    ----------
    ratio : numbers.Real, default 6
        Mean ratio of the stretch or compress.
        If it is in the interval [1, 100],
        then it will be transformed to [0, 1].
        The ratio of one batch element is sampled from a normal distribution.
    prob : float, default 0.5
        Probability of the augmenter to be applied.
    overlap : float, default 0.5
        Overlap of offline generated data.
    critical_overlap : float, default 0.85
        Overlap of the critical region of the ECG.

    Example
    -------
    .. code-block:: python

        sco = StretchCompressOffline()
        seglen = 600
        sig = torch.randn((12, 60000)).numpy()
        labels = torch.ones((60000, 3)).numpy().astype(int)
        masks = torch.ones((60000, 1)).numpy().astype(int)
        segments = sco(600, sig, labels, masks, critical_points=[10000,30000])

    """

    __name__ = "StretchCompressOffline"

    def __init__(
        self,
        ratio: Real = 6,
        prob: float = 0.5,
        overlap: float = 0.5,
        critical_overlap: float = 0.85,
    ) -> None:
        self.prob = prob
        assert 0 <= self.prob <= 1, "Probability must be between 0 and 1"
        self.ratio = ratio
        if self.ratio > 1:
            self.ratio = self.ratio / 100
        assert (
            0 <= self.ratio <= 1
        ), "Ratio must be between 0 and 1, or between 0 and 100"
        self.overlap = overlap
        assert (
            0 <= self.overlap < 1
        ), "Overlap ratio must be between 0 and 1 (1 not included)"
        self.critical_overlap = critical_overlap
        assert (
            0 <= self.critical_overlap < 1
        ), "Critical overlap ratio must be between 0 and 1 (1 not included)"

    def generate(
        self,
        seglen: int,
        sig: np.ndarray,
        *labels: Sequence[np.ndarray],
        critical_points: Optional[Sequence[int]] = None,
    ) -> List[Tuple[Union[np.ndarray, int], ...]]:
        """Generate stretched or compressed segments from the ECGs.

        Parameters
        ----------
        seglen : int
            Length of the ECG segments to be generated.
        sig : numpy.ndarray,
            THe ECGs to generate stretched or compressed segments,
            of shape ``(lead, siglen)``.
        labels : numpy.ndarray, optional
            Labels of the ECGs, of shape ``(label_len, channels)``.
            For example, when doing segmentation,
            `label_len` should be divisible by `siglen`,
            `channels` should be the same as the number of classes.
        critical_points : Sequence[int], optional
            Indices of the critical points of the ECG,
            usually have larger overlap by :attr:`self.critical_overlap`.

        Returns
        -------
        list
            list of generated segments,
            consisting segments of the form
            ``(seg, label1, label2, ..., start_idx, end_idx)``.

        """
        siglen = sig.shape[1]
        forward_len = int(round(seglen - seglen * self.overlap))
        critical_forward_len = int(round(seglen - seglen * self.critical_overlap))
        critical_forward_len = [critical_forward_len // 4, critical_forward_len]
        # print(forward_len, critical_forward_len)

        # skip those records that are too short
        if siglen < seglen:
            return []

        segments = []

        # ordinary segments with constant forward_len
        for idx in range((siglen - seglen) // forward_len + 1):
            start_idx = idx * forward_len
            new_seg = self.__generate_segment(
                seglen,
                sig,
                *labels,
                start_idx=start_idx,
            )
            segments.append(new_seg)
        # the tail segment
        if (siglen - seglen) % forward_len != 0:
            new_seg = self.__generate_segment(
                seglen,
                sig,
                *labels,
                end_idx=siglen,
            )
            segments.append(new_seg)

        # special segments around critical_points with random forward_len in critical_forward_len
        for cp in critical_points or []:
            start_idx = max(
                0,
                cp - seglen + randint(critical_forward_len[0], critical_forward_len[1]),
            )
            while start_idx <= min(cp - critical_forward_len[1], siglen - seglen):
                new_seg = self.__generate_segment(
                    seglen,
                    sig,
                    *labels,
                    start_idx=start_idx,
                )
                segments.append(new_seg)
                start_idx += randint(critical_forward_len[0], critical_forward_len[1])
        return segments

    def __generate_segment(
        self,
        seglen: int,
        sig: np.ndarray,
        *labels: Sequence[np.ndarray],
        start_idx: Optional[int] = None,
        end_idx: Optional[int] = None,
    ) -> Tuple[Union[np.ndarray, int], ...]:
        """Internal function to generate a stretched or compressed segment.

        Parameters
        ----------
        seglen : int
            Length of the ECG segments to be generated.
        sig : numpy.ndarray
            ECGs to generate stretched or compressed segments,
            of shape ``(lead, siglen)``.
        labels : numpy.ndarray, optional
            lLbels of the ECGs, of shape ``(label_len, channels)``.
            For example, when doing segmentation,
            `label_len` should be divisible by `siglen`,
            `channels` should be the same as the number of classes.
        start_idx : int, optional
            Start index of the segment in `sig`.
        end_idx : int, optional
            End index of the segment in `sig`.
            If `start_idx` is set, then `end_idx` is ignored.
            At least one of `start_idx` and `end_idx` should be set.

        Returns
        -------
        tuple
            Tuple of generated segment,
            consisting of segments of the form
            ``(seg, label1, label2, ..., start_idx, end_idx)``.

        """
        assert not all(
            [start_idx is None, end_idx is None]
        ), "at least one of `start_idx` and `end_idx` should be set"

        siglen = sig.shape[1]
        ratio = self._sample_ratio()
        aug_labels = []
        if ratio != 0:
            sign = choice([-1, 1])
            new_len = int(round((1 + sign * ratio) * seglen))
            if start_idx is not None:
                end_idx = start_idx + new_len
            else:
                start_idx = end_idx - new_len
            if end_idx > siglen:
                end_idx = siglen
                start_idx = max(0, end_idx - new_len)
                ratio = (end_idx - start_idx) / seglen - 1
            aug_seg = sig[..., start_idx:end_idx]
            aug_seg = SS.resample(x=aug_seg, num=seglen, axis=1)
            for lb in labels:
                dtype = lb.dtype
                aug_labels.append(
                    F.interpolate(
                        torch.from_numpy(
                            lb[start_idx:end_idx, ...].T.astype(np.float32)
                        ).unsqueeze(0),
                        size=seglen,
                        mode="nearest",
                    )
                    .squeeze(0)
                    .numpy()
                    .T.astype(dtype)
                )
        else:
            if start_idx is not None:
                end_idx = start_idx + seglen
                if end_idx > siglen:
                    end_idx = siglen
                    start_idx = end_idx - seglen
            else:
                start_idx = end_idx - seglen
                if start_idx < 0:
                    start_idx = 0
                    end_idx = seglen
            aug_seg = sig[..., start_idx:end_idx]
            for lb in labels:
                aug_labels.append(lb[start_idx:end_idx, ...])
        return (aug_seg,) + tuple(aug_labels) + (start_idx, end_idx)

    def _sample_ratio(self) -> float:
        """Sample the ratio of stretching or compressing."""
        if DEFAULTS.RNG.uniform() >= self.prob:
            return 0
        else:
            return np.clip(
                DEFAULTS.RNG.normal(self.ratio, 0.382 * self.ratio),
                0.01 * self.ratio,
                2 * self.ratio,
            )

    @add_docstring(generate.__doc__)
    def __call__(
        self,
        seglen: int,
        sig: np.ndarray,
        *labels: Sequence[np.ndarray],
        critical_points: Optional[Sequence[int]] = None,
    ) -> List[Tuple[np.ndarray, ...]]:
        return self.generate(seglen, sig, *labels, critical_points=critical_points)

    def extra_repr_keys(self) -> List[str]:
        return super().extra_repr_keys() + [
            "ratio",
            "prob",
            "overlap",
            "critical_overlap",
        ]



"""
Sequence labeling nets, for wave delineation,

the labeling granularity is the frequency of the input signal,
divided by the length (counted by the number of basic blocks) of each branch

Pipeline
--------
multi-scopic cnn --> (bidi-lstm -->) "attention" (se block) --> seq linear

References
----------
[1] Cai, Wenjie, and Danqin Hu. "QRS complex detection using novel deep learning neural networks." IEEE Access (2020).

"""

import warnings
from copy import deepcopy
from typing import Optional, Sequence, Union, List

import torch
import torch.nn.functional as F
from torch import Tensor

from .ecg_crnn import ECG_CRNN, ECG_CRNN_v1
from ..model_configs.ecg_seq_lab_net import ECG_SEQ_LAB_NET_CONFIG
from ..cfg import CFG
from ..utils import add_docstring


__all__ = [
    "ECG_SEQ_LAB_NET",
]


class ECG_SEQ_LAB_NET(ECG_CRNN):
    """SOTA model from CPSC2019 challenge.

    Sequence labeling nets, for wave delineation, QRS complex detection, etc.
    Proposed in [1]_.

    pipeline
    --------
    (multi-scopic, etc.) cnn --> head ((bidi-lstm -->) "attention" --> seq linear) -> output


    Parameters
    ----------
    classes : List[str]
        List of the classes for sequence labeling.
    n_leads : int
        Number of leads (number of input channels).
    config : dict, optional
        Other hyper-parameters, including kernel sizes, etc.
        Refer to corresponding config file.

    References
    ----------
    .. [1] Cai, Wenjie, and Danqin Hu. "QRS complex detection using novel deep learning neural networks." IEEE Access (2020).

    """

    __name__ = "ECG_SEQ_LAB_NET"
    __DEFAULT_CONFIG__ = {"recover_length": False}
    __DEFAULT_CONFIG__.update(deepcopy(ECG_SEQ_LAB_NET_CONFIG))

    def __init__(
        self, classes: Sequence[str], n_leads: int, config: Optional[CFG] = None
    ) -> None:
        _config = CFG(deepcopy(self.__DEFAULT_CONFIG__))
        if not config:
            warnings.warn(
                "No config is provided, using default config.", RuntimeWarning
            )
        _config.update(deepcopy(config) or {})
        _config.global_pool = "none"
        super().__init__(classes, n_leads, _config)

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, channels, seq_len)``.

        Returns
        -------
        pred : torch.Tensor
            Output tensor,
            of shape ``(batch_size, seq_len, n_classes)``

        """
        batch_size, channels, seq_len = input.shape

        pred = super().forward(input)  # (batch_size, len, n_classes)

        if self.config.recover_length:
            pred = F.interpolate(
                pred.permute(0, 2, 1),
                size=seq_len,
                mode="linear",
                align_corners=True,
            ).permute(0, 2, 1)

        return pred

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the model.

        Parameters
        ----------
        seq_len : int, optional
            Length of the 1d input signal tensor.
        batch_size : int, optional
            Batch size of the input signal tensor.

        Returns
        -------
        output_shape : sequence
            The output shape of the model.

        """
        output_shape = super().compute_output_shape(seq_len, batch_size)
        if self.config.recover_length:
            output_shape = (batch_size, seq_len, output_shape[-1])
        return output_shape

    @classmethod
    def from_v1(
        cls, v1_ckpt: str, device: Optional[torch.device] = None
    ) -> "ECG_SEQ_LAB_NET":
        """Convert the v1 model to the current version.

        Parameters
        ----------
        v1_ckpt : str
            Path to the v1 checkpoint file.

        Returns
        -------
        model : ECG_SEQ_LAB_NET
            The converted model.

        """
        v1_model, _ = ECG_SEQ_LAB_NET_v1.from_checkpoint(v1_ckpt, device=device)
        model = cls(
            classes=v1_model.classes, n_leads=v1_model.n_leads, config=v1_model.config
        )
        model = model.to(v1_model.device)
        model.cnn.load_state_dict(v1_model.cnn.state_dict())
        if model.rnn.__class__.__name__ != "Identity":
            model.rnn.load_state_dict(v1_model.rnn.state_dict())
        if model.attn.__class__.__name__ != "Identity":
            model.attn.load_state_dict(v1_model.attn.state_dict())
        model.clf.load_state_dict(v1_model.clf.state_dict())
        del v1_model
        return model

    @property
    def doi(self) -> List[str]:
        return list(set(super().doi + ["10.1109/access.2020.2997473"]))


@add_docstring(ECG_SEQ_LAB_NET.__doc__)
class ECG_SEQ_LAB_NET_v1(ECG_CRNN_v1):

    __name__ = "ECG_SEQ_LAB_NET_v1"
    __DEFAULT_CONFIG__ = {"recover_length": False}
    __DEFAULT_CONFIG__.update(deepcopy(ECG_SEQ_LAB_NET_CONFIG))

    def __init__(
        self, classes: Sequence[str], n_leads: int, config: Optional[CFG] = None
    ) -> None:
        _config = CFG(deepcopy(self.__DEFAULT_CONFIG__))
        if not config:
            warnings.warn(
                "No config is provided, using default config.", RuntimeWarning
            )
        _config.update(deepcopy(config) or {})
        _config.global_pool = "none"
        super().__init__(classes, n_leads, _config)

    def extract_features(self, input: Tensor) -> Tensor:
        """Extract feature map before the dense (linear) classifying layer(s).

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, channels, seq_len)``.

        Returns
        -------
        features : torch.Tensor
            Feature map tensor,
            of shape ``(batch_size, seq_len, channels)``.

        """
        # cnn
        cnn_output = self.cnn(input)  # (batch_size, channels, seq_len)

        # rnn or none
        if self.rnn:
            rnn_output = cnn_output.permute(2, 0, 1)  # (seq_len, batch_size, channels)
            rnn_output = self.rnn(rnn_output)  # (seq_len, batch_size, channels)
            rnn_output = rnn_output.permute(1, 2, 0)  # (batch_size, channels, seq_len)
        else:
            rnn_output = cnn_output

        # attention
        if self.attn:
            features = self.attn(rnn_output)  # (batch_size, channels, seq_len)
        else:
            features = rnn_output
        # features = features.permute(0, 2, 1)  # (batch_size, seq_len, channels)
        return features

    @add_docstring(ECG_SEQ_LAB_NET.forward.__doc__)
    def forward(self, input: Tensor) -> Tensor:
        batch_size, channels, seq_len = input.shape

        pred = super().forward(input)  # (batch_size, len, n_classes)

        if self.config.recover_length:
            pred = F.interpolate(
                pred.permute(0, 2, 1),
                size=seq_len,
                mode="linear",
                align_corners=True,
            ).permute(0, 2, 1)

        return pred

    @property
    def doi(self) -> List[str]:
        return list(set(super().doi + ["10.1109/access.2020.2997473"]))



"""Normalization of the signals."""

from numbers import Real
from typing import Any, List, Tuple, Union

import numpy as np

from ..utils.utils_signal import normalize
from .base import PreProcessor

__all__ = [
    "Normalize",
    "MinMaxNormalize",
    "NaiveNormalize",
    "ZScoreNormalize",
]


class Normalize(PreProcessor):
    r"""Normalization of the signals.

    Perform z-score normalization on `sig`,
    to make it has fixed mean and standard deviation;
    or perform min-max normalization on `sig`,
    or normalize `sig` using `mean` and `std` via
    :math:`(sig - mean) / std`.
    More precisely,

    .. math::

        \begin{align*}
        \text{Min-Max normalization:} & \quad \frac{sig - \min(sig)}{\max(sig) - \min(sig)} \\
        \text{Naive normalization:} & \quad \frac{sig - m}{s} \\
        \text{Z-score normalization:} & \quad \left(\frac{sig - mean(sig)}{std(sig)}\right) \cdot s + m
        \end{align*}

    Parameters
    ----------
    method : {"naive", "min-max", "z-score"}, optional
        Normalization method, case insensitive,
        default "z-score".
    mean : numbers.Real or numpy.ndarray, default 0.0
        Mean value of the normalized signal,
        or mean values for each lead of the normalized signal.
        Useless if `method` is "min-max".
    std : numbers.Real or numpy.ndarray, default 1.0
        Standard deviation of the normalized signal,
        or standard deviations for each lead of the normalized signal.
        Useless if `method` is "min-max".
    per_channel : bool, default False
        If True, normalization will be done per channel.

    Examples
    --------
    .. code-block:: python

        from torch_ecg.cfg import DEFAULTS
        sig = DEFAULTS.RNG.randn(1000)
        pp = Normalize(method="z-score", mean=0.0, std=1.0)
        sig, _ = pp(sig, 500)

    """

    __name__ = "Normalize"

    def __init__(
        self,
        method: str = "z-score",
        mean: Union[Real, np.ndarray] = 0.0,
        std: Union[Real, np.ndarray] = 1.0,
        per_channel: bool = False,
        **kwargs: Any,
    ) -> None:
        self.method = method.lower()
        assert self.method in [
            "z-score",
            "naive",
            "min-max",
        ]
        self.mean = mean
        self.std = std
        self.per_channel = per_channel
        if isinstance(std, Real):
            assert std > 0, "standard deviation should be positive"
        else:
            assert (std > 0).all(), "standard deviations should all be positive"
        if not per_channel:
            assert isinstance(mean, Real) and isinstance(
                std, Real
            ), "mean and std should be real numbers in the non per-channel setting"

    def apply(self, sig: np.ndarray, fs: Real) -> Tuple[np.ndarray, int]:
        """Apply the preprocessor to `sig`.

        Parameters
        ----------
        sig : numpy.ndarray
            The ECG signal, can be
                - 1d array, which is a single-lead ECG;
                - 2d array, which is a multi-lead ECG of "lead_first" format;
                - 3d array, which is a tensor of several ECGs, of shape ``(batch, lead, siglen)``.
        fs : numbers.Real
            Sampling frequency of the ECG signal.
            **NOT** used currently.

        Returns
        -------
        normalized_sig : :class:`numpy.ndarray`
            The normalized ECG signal.
        fs : :class:`int`
            The sampling frequency of the normalized ECG signal.

        """
        self._check_sig(sig)
        normalized_sig = normalize(
            sig=sig,
            method=self.method,
            mean=self.mean,
            std=self.std,
            sig_fmt="channel_first",
            per_channel=self.per_channel,
        )
        return normalized_sig, fs

    def extra_repr_keys(self) -> List[str]:
        return [
            "method",
            "mean",
            "std",
            "per_channel",
        ] + super().extra_repr_keys()


class MinMaxNormalize(Normalize):
    r"""Min-Max normalization.

    Min-Max normalization defined as

    .. math::

        \frac{sig - \min(sig)}{\max(sig) - \min(sig)}

    Parameters
    ----------
    per_channel : bool, default False
        If True, normalization will be done per channel.

    Examples
    --------
    .. code-block:: python

        from torch_ecg.cfg import DEFAULTS
        sig = DEFAULTS.RNG.randn(1000)
        pp = MinMaxNormalize()
        sig, _ = pp(sig, 500)

    """

    __name__ = "MinMaxNormalize"

    def __init__(
        self,
        per_channel: bool = False,
    ) -> None:
        super().__init__(method="min-max", per_channel=per_channel)

    def extra_repr_keys(self) -> List[str]:
        return [
            "per_channel",
        ] + super().extra_repr_keys()


class NaiveNormalize(Normalize):
    r"""Naive normalization.

    Naive normalization defined as

    .. math::

        \frac{sig - m}{s}

    Parameters
    ----------
    mean : numbers.Real or numpy.ndarray, default 0.0
        Value(s) to be subtracted.
    std : numbers.Real or numpy.ndarray, default 1.0
        Value(s) to be divided.
    per_channel : bool, default False
        If True, normalization will be done per channel.

    Examples
    --------
    .. code-block:: python

        from torch_ecg.cfg import DEFAULTS
        sig = DEFAULTS.RNG.randn(1000)
        pp = NaiveNormalize()
        sig, _ = pp(sig, 500)

    """

    __name__ = "NaiveNormalize"

    def __init__(
        self,
        mean: Union[Real, np.ndarray] = 0.0,
        std: Union[Real, np.ndarray] = 1.0,
        per_channel: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            method="naive",
            mean=mean,
            std=std,
            per_channel=per_channel,
        )

    def extra_repr_keys(self) -> List[str]:
        return [
            "mean",
            "std",
            "per_channel",
        ] + super().extra_repr_keys()


class ZScoreNormalize(Normalize):
    r"""Z-score normalization.

    Z-score normalization defined as

    .. math::

        \left(\frac{sig - mean(sig)}{std(sig)}\right) \cdot s + m

    Parameters
    ----------
    mean : numbers.Real or numpy.ndarray, default 0.0
        Mean value of the normalized signal,
        or mean values for each lead of the normalized signal.
    std : numbers.Real or numpy.ndarray, default 1.0
        Standard deviation of the normalized signal,
        or standard deviations for each lead of the normalized signal.
    per_channel : bool, default False
        If True, normalization will be done per channel.

    Examples
    --------
    .. code-block:: python

        from torch_ecg.cfg import DEFAULTS
        sig = DEFAULTS.RNG.randn(1000)
        pp = ZScoreNormalize()
        sig, _ = pp(sig, 500)

    """

    __name__ = "ZScoreNormalize"

    def __init__(
        self,
        mean: Union[Real, np.ndarray] = 0.0,
        std: Union[Real, np.ndarray] = 1.0,
        per_channel: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            method="z-score",
            mean=mean,
            std=std,
            per_channel=per_channel,
        )

    def extra_repr_keys(self) -> List[str]:
        return [
            "mean",
            "std",
            "per_channel",
        ] + super().extra_repr_keys()
